#include <stdio.h>
#include <stdlib.h>

#define MAX 200
#define NUM 10

double probs[NUM] = {.26, .47, .63, .76, .84, .90, .94, .97, .99, 1.0};
int    slots[NUM] = { 10,  20,  30,  40,  50,  60,  70,  80,  90, 100};

main(void)
{
   /*
   **   Listing8.1.c - Generate random delay times from IAT observations
   */
   double random, delay;
   double dtab[NUM][2]; /* probabilities and cumulative probabilities */
   int ii, jj, kk;
   int itab[NUM];       /* counts of observations within range slots */
   int obs[MAX];        /* observations */

   srand((unsigned int) getpid()); /* seed rand() */

   /*
   **   Generate delay times
   */
   for(ii=0; ii<MAX; ii++)
   {
      /* get values between 0.0 and 1.0 */
      random = rand() / (double) RAND_MAX;

      /* match a time slot */
      for(jj=0; jj<NUM; jj++)
      {
         if(random <= probs[jj])
            break;
      }
      /* random values within each slot */
      random = rand() / (double) RAND_MAX;
      delay = (double) slots[jj] - (random * (double) slots[0]);

      /********************************************************/
      /*** send each query after the calculated delay times ***/
      /********************************************************/

      /* save delay times */
      obs[ii] = (int) delay;
      if((delay - (double) obs[ii]) > 0.5L)
         obs[ii]++; /* round up if necessary */
      if(obs[ii] == 0)
         obs[ii]++; /* no simultaneous events */
   }

   /*
   **   Analyze observed (generated) delay times
   */
   for(ii=0; ii<NUM; ii++)
      itab[ii] = 0; /* zero counters */
   for(ii=0; ii<MAX; ii++)
   {
      for(jj=0; jj<NUM-1; jj++)
         if(slots[jj] > obs[ii])
            break;
      itab[jj]++;
   }

   /* calc probabilities */
   for(ii=0; ii<NUM; ii++)
      dtab[ii][0] = ((double) itab[ii] / (double) MAX);

   /* cumm probabilities */
   dtab[0][1] = dtab[0][0];
   for(ii=1; ii<NUM; ii++)
      dtab[ii][1] = dtab[ii-1][1] + dtab[ii][0];

   /* print out the table */
   printf("\n\t  RANGE   COUNT   PROB   CUMM\n");
   printf("\t ------   -----   ----   ----\n");
   for(ii=0; ii<NUM; ii++)
      printf("\t %2d-%3d   %5d   %4.2lf   %4.2lf\n",
         (ii*NUM)+1, ((ii+1)*NUM), itab[ii],
         dtab[ii][0], dtab[ii][1]);
   printf("\n");
}
